/* --COPYRIGHT--,BSD
 * Copyright (c) 2010, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xdc/std.h>

#include <ti/sdo/dmai/Dmai.h>
#include <ti/sdo/dmai/VideoStd.h>
#include <ti/sdo/dmai/Display.h>

#include "priv/_SysFs.h"

#define MODULE_NAME     "Dmai"

/* Macro for calling system */
#define csys(x) \
    ret = system((x)); \
    if (ret < 0) return Dmai_EINVAL; \
    if (ret != 0) return Dmai_EFAIL

/* Support for LCD & DVI only */

/* fbdev Device commands */
typedef enum {
    FsOptions_COMMANDS_END = 0,
    FsOptions_DISABLE_DISPLAY,
    FsOptions_ENABLE_DISPLAY,
    FsOptions_MANAGER_NONE,
    FsOptions_MANAGER_LCD,
    FsOptions_MANAGER_DVI,
    FsOptions_MANAGER_TV,
    FsOptions_FBSET_LCD,
    FsOptions_FBSET_TV_LCD,
    FsOptions_FBSET_DVI,
    FsOptions_FBSET_TV_DVI,
    FsOptions_FBSET_TV
} FsOptions;

typedef struct FsCommand {
    FsOptions    command;
    Char        *value;
} FsCommand;

typedef struct FsOptionsSequence {
    FsOptions    command;
    Char        *sysFB;
} FsOptionsSequence;

const struct FsCommand displayCommand[] = {
    { FsOptions_DISABLE_DISPLAY,    "0"},
    { FsOptions_ENABLE_DISPLAY,     "1"},
    { FsOptions_MANAGER_NONE,       ""},
    { FsOptions_MANAGER_TV,         "tv"},
    { FsOptions_MANAGER_DVI,        "dvi"},
    { FsOptions_MANAGER_LCD,        "lcd"},
    { FsOptions_FBSET_LCD,          "fbset lcd"},
    { FsOptions_FBSET_TV_LCD,       "fbset lcd"},
    { FsOptions_FBSET_DVI,          "fbset dvi"},
    { FsOptions_FBSET_TV_DVI,       "fbset dvi"},
    { FsOptions_FBSET_TV,           "fbset tv"},
    { 0,                            NULL}
};

#define SETFORMAT_FROM_LCD_DVI \
    "fbset -fb /dev/fb0 -xres 480 -yres 720 -vxres 480 -vyres 720"
#define SETFORMAT_FROM_DVI_LCD \
    "fbset -fb /dev/fb0 -xres 480 -yres 640 -vxres 480 -vyres 640"
#define SYSFS_MANAGER0_DISPLAY "/sys/devices/platform/omapdss/manager0/display"
#define SYSFS_OVERLAY0_DISPLAY "/sys/devices/platform/omapdss/overlay0/manager"
#define SYSFS_OVERLAY1_DISPLAY "/sys/devices/platform/omapdss/overlay1/manager"
#define SYSFS_OVERLAY2_DISPLAY "/sys/devices/platform/omapdss/overlay2/manager"
#define SYSFS_OVERLAY0_ENABLED "/sys/devices/platform/omapdss/overlay0/enabled"
#define SYSFS_DISPLAY0_ENABLED "/sys/devices/platform/omapdss/display0/enabled"
#define SYSFS_DISPLAY1_ENABLED "/sys/devices/platform/omapdss/display1/enabled"
#define SYSFS_DISPLAY2_ENABLED "/sys/devices/platform/omapdss/display2/enabled"

#define READ_BUFFER_SIZE 100

/******************************************************************************
 * writeSysFs
 ******************************************************************************/
static Int writeSysFs(Char *fileName, Char *val)
{
    FILE *fp;
    Char *valString;

    valString = malloc(strlen(val) + 1);

    if (valString == NULL) {
        Dmai_err0("Failed to allocate memory for temporary string\n");
        return Dmai_ENOMEM;
    }

    fp = fopen(fileName, "w");

    if (fp == NULL) {
        Dmai_err1("Failed to open %s for writing\n", fileName);
        free(valString);
        return Dmai_EIO;
    }

    if (fwrite(val, strlen(val) + 1, 1, fp) != 1) {
        Dmai_err2("Failed to write sysfs variable %s to %s\n",
                  fileName, val);
        fclose(fp);
        free(valString);
        return Dmai_EIO;
    }

    fclose(fp);

    free(valString);

    return Dmai_EOK;
}

/******************************************************************************
 * readSysFs
 ******************************************************************************/
static Int readSysFs(Char *fileName, Char *val, int length)
{
    FILE *fp;
    int ret;
    int len;
    char *tok;

    fp = fopen(fileName, "r");

    if (fp == NULL) {
        Dmai_err1("Failed to open %s for reading\n", fileName);
        return Dmai_EIO;
    }

    memset(val, '\0', length);

    ret = fread(val, 1, length, fp);

    if (ret < 1) {
        Dmai_err1("Failed to read sysfs variable from %s\n", fileName);
        return Dmai_EIO;
    }

    tok = strtok(val, "\n");
    len = tok ? strlen(tok) : strlen(val);
    val[len] = '\0';

    fclose(fp);

    return Dmai_EOK;
}

/******************************************************************************
 * detectVideoType
 * TODO : The current implementation is checking the Display type and setting
 * video type.Implementation to check the setting from SYSFS setting is pending.
 *****************************************************************************/
static Int detectVideoType(VideoStd_Type *videoStdType)
{
    Char videoStd[READ_BUFFER_SIZE];

    /* Read Displays in sysfs file */
    if (readSysFs(SYSFS_MANAGER0_DISPLAY, videoStd, 
                  READ_BUFFER_SIZE) < 0) {
        Dmai_err1("Failed to read %s\n", SYSFS_MANAGER0_DISPLAY);
        return Dmai_EIO;
    }

    if (!strcmp(videoStd, "lcd")){
        *videoStdType = VideoStd_VGA;
    } else if (!strcmp(videoStd, "dvi")){
        *videoStdType = VideoStd_480P;
    } else if (!strcmp(videoStd, "tv")){
        *videoStdType = VideoStd_D1_NTSC;
    } else {
        Dmai_err0("Unknown Device Output Found\n");
        return Dmai_EFAIL;
    }

    return Dmai_EOK;
}

/******************************************************************************
 * readOutputDevice
 *****************************************************************************/
static Int readOutputDevice(Display_Output *displayOutput)
{
    Char device[READ_BUFFER_SIZE];

    /* Read overlays in sysfs file */
    if (readSysFs(SYSFS_MANAGER0_DISPLAY, device,
                        READ_BUFFER_SIZE) < 0) {
        Dmai_err1("Failed to read %s\n", SYSFS_MANAGER0_DISPLAY);
        return Dmai_EIO;
    }           

    if (!strcmp(device, "lcd")) {
        *displayOutput = Display_Output_LCD;        

        if (readSysFs(SYSFS_OVERLAY0_DISPLAY, device,
                            READ_BUFFER_SIZE) < 0) {
            Dmai_err1("Failed to read %s\n", SYSFS_OVERLAY0_DISPLAY);
            return Dmai_EIO;
        }
        if (!strcmp(device, "tv")) {
            *displayOutput = Display_Output_SVIDEO;
        }
    } else if (!strcmp(device, "dvi")) {
        *displayOutput = Display_Output_DVI;

        if (readSysFs(SYSFS_OVERLAY0_DISPLAY, device,
                      READ_BUFFER_SIZE) < 0) {
            Dmai_err1("Failed to read %s\n", SYSFS_OVERLAY0_DISPLAY);
            return Dmai_EIO;
        }
        if (!strcmp(device, "tv")) {
            *displayOutput = Display_Output_SVIDEO;
        }
    } else {
        Dmai_err0("Unknown Device Output Found\n");
        return Dmai_EFAIL;
    }           

    return Dmai_EOK;
}

/******************************************************************************
 * setupDevice
 *****************************************************************************/
static Int setupDevice(FsOptions command, Char *fileName)
{
    int index = 0;
    int ret;

    while ((command != displayCommand[index].command) && 
           (displayCommand[index].command != FsOptions_COMMANDS_END)) {
        index++;
    }

    if (!displayCommand[index].command) {
        Dmai_err0("Unsupported command\n");
        return Dmai_EINVAL;
    }

    if (displayCommand[index].command == FsOptions_FBSET_DVI) {
        csys("echo \"\" > /sys/devices/platform/omapdss/manager0/display");
        csys("fbset -fb /dev/fb0 -xres 720 -yres 480 -vxres 720 -vyres 480");
    }
    else if (displayCommand[index].command == FsOptions_FBSET_LCD) {
        csys("echo \"\" > /sys/devices/platform/omapdss/manager0/display");
        csys("fbset -fb /dev/fb0 -xres 480 -yres 640 -vxres 480 -vyres 640");
    }
    else if (displayCommand[index].command == FsOptions_FBSET_TV) {
        csys("echo 0 > /sys/devices/platform/omapdss/overlay0/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display0/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display1/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display2/enabled");
        csys("echo \"\" > /sys/devices/platform/omapdss/manager0/display");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay0/manager");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay1/manager");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay2/manager");
        csys("fbset -fb /dev/fb0 -xres 720 -yres 480 -vxres 720 -vyres 480 -depth 16");
    }
    else if (displayCommand[index].command == FsOptions_FBSET_TV_LCD) {
        csys("echo 0 > /sys/devices/platform/omapdss/overlay0/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display0/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display1/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display2/enabled");
        csys("echo \"\" > /sys/devices/platform/omapdss/manager0/display");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay0/manager");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay1/manager");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay2/manager");
        csys("fbset -fb /dev/fb0 -xres 480 -yres 640 -vxres 480 -vyres 640 ");
    }
    else if (displayCommand[index].command == FsOptions_FBSET_TV_DVI) {
        csys("echo 0 > /sys/devices/platform/omapdss/overlay0/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display0/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display1/enabled");
        csys("echo 0 > /sys/devices/platform/omapdss/display2/enabled");
        csys("echo \"\" > /sys/devices/platform/omapdss/manager0/display");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay0/manager");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay1/manager");
        csys("echo \"\" > /sys/devices/platform/omapdss/overlay2/manager");
        csys("fbset -fb /dev/fb0 -xres 720 -yres 480 -vxres 720 -vyres 480");
    }
    else if(writeSysFs(fileName, displayCommand[index].value) < 0) {
        Dmai_err2("Failed to write %s on %s\n", displayCommand[index].value, 
            fileName);
        return Dmai_EIO;
    }

    return Dmai_EOK;
}

/******************************************************************************
 * _SysFs_Change (INTERNAL)
 ******************************************************************************/
Int _SysFs_change(Display_Output *displayOutput, Char* displayDevice,
                  VideoStd_Type *videoType, Int *rotation)
{
    FsOptionsSequence *sequence = NULL;
    Display_Output currentDisplayOutput;

    /* List of Sequences using the FileSystem entries to switch displays */        
    const FsOptionsSequence sequence_LCD_DVI[] = {
        { FsOptions_DISABLE_DISPLAY,     SYSFS_DISPLAY0_ENABLED },
        { FsOptions_FBSET_DVI,           NULL },
        { FsOptions_MANAGER_DVI,         SYSFS_MANAGER0_DISPLAY },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_DISPLAY2_ENABLED },
        { FsOptions_COMMANDS_END,        NULL }
    };
    const FsOptionsSequence sequence_LCD_SVIDEO[] = {
        { FsOptions_FBSET_TV,            NULL },
        { FsOptions_MANAGER_TV,          SYSFS_OVERLAY0_DISPLAY },
        { FsOptions_MANAGER_TV,          SYSFS_OVERLAY1_DISPLAY },
        { FsOptions_MANAGER_TV,          SYSFS_OVERLAY2_DISPLAY },
        { FsOptions_MANAGER_LCD,         SYSFS_MANAGER0_DISPLAY },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_OVERLAY0_ENABLED },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_DISPLAY1_ENABLED },
        { FsOptions_COMMANDS_END,        NULL }
    };
    const FsOptionsSequence sequence_DVI_LCD[] = {
        { FsOptions_DISABLE_DISPLAY,     SYSFS_DISPLAY2_ENABLED },
        { FsOptions_FBSET_LCD,           NULL },
        { FsOptions_MANAGER_LCD,         SYSFS_MANAGER0_DISPLAY },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_DISPLAY0_ENABLED },
        { FsOptions_COMMANDS_END,        NULL }
    };
    const FsOptionsSequence sequence_DVI_SVIDEO[] = {
        { FsOptions_FBSET_TV,            NULL },
        { FsOptions_MANAGER_TV,          SYSFS_OVERLAY0_DISPLAY },
        { FsOptions_MANAGER_TV,          SYSFS_OVERLAY1_DISPLAY },
        { FsOptions_MANAGER_TV,          SYSFS_OVERLAY2_DISPLAY },
        { FsOptions_MANAGER_DVI,         SYSFS_MANAGER0_DISPLAY },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_OVERLAY0_ENABLED },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_DISPLAY1_ENABLED },
        { FsOptions_COMMANDS_END,        NULL }
    };
    const FsOptionsSequence sequence_SVIDEO_LCD[] = {
        { FsOptions_FBSET_TV_LCD,        NULL },
        { FsOptions_MANAGER_LCD,         SYSFS_OVERLAY0_DISPLAY },
        { FsOptions_MANAGER_LCD,         SYSFS_OVERLAY1_DISPLAY },
        { FsOptions_MANAGER_LCD,         SYSFS_OVERLAY2_DISPLAY },
        { FsOptions_MANAGER_LCD,         SYSFS_MANAGER0_DISPLAY },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_OVERLAY0_ENABLED },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_DISPLAY0_ENABLED },
        { FsOptions_COMMANDS_END,        NULL }
    };
    const FsOptionsSequence sequence_SVIDEO_DVI[] = {
        { FsOptions_FBSET_TV_DVI,        NULL },
        { FsOptions_MANAGER_LCD,         SYSFS_OVERLAY0_DISPLAY },
        { FsOptions_MANAGER_LCD,         SYSFS_OVERLAY1_DISPLAY },
        { FsOptions_MANAGER_LCD,         SYSFS_OVERLAY2_DISPLAY },
        { FsOptions_MANAGER_DVI,         SYSFS_MANAGER0_DISPLAY },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_OVERLAY0_ENABLED },
        { FsOptions_ENABLE_DISPLAY,      SYSFS_DISPLAY2_ENABLED },
        { FsOptions_COMMANDS_END,        NULL }
    };

    /* Identify from the current FS settings the display that is set */ 
    if (readOutputDevice(&currentDisplayOutput) < 0) {
        Dmai_err0("Failed to Detect Current Display Output.\n");
        Dmai_err1("Attempting to change to %d\n", *displayOutput);
        return Dmai_EIO;
    }

    /* Don't make any changes when SYSTEM is used */
    if (*displayOutput == Display_Output_SYSTEM) {
        *displayOutput = currentDisplayOutput; 
        /*
         * Identify the appropriate Video Type dimensions based on the
         * sys fs settings.
         */
        if (*videoType == VideoStd_AUTO) {
            if (detectVideoType(videoType) < 0) {
                Dmai_err0("Failed to Detect Current VideoType\n");
                return Dmai_EIO;
            }      
        }

        return Dmai_EOK;
    }

    /* Are we using the display that was selected? */
    if (*displayOutput == currentDisplayOutput) {
        /* No need to make the transition */
        return Dmai_EOK;
    }

    /* Debugging code to show the display change */
    if (*displayOutput != Display_Output_SYSTEM) {
        Dmai_dbg2("Current Display Output set as %d, switching to %d.\n", 
                  currentDisplayOutput, *displayOutput);
    }

    /*
     * Identify the appropriate sequence depending of the current and
     * desired Display settings. Currently DVI, LCD and TV are only assumed.
     */
    switch(*displayOutput) {
        case Display_Output_LCD:
             if (currentDisplayOutput == Display_Output_DVI) {
                 sequence = (FsOptionsSequence *) sequence_DVI_LCD;
             }
             else if (currentDisplayOutput == Display_Output_SVIDEO) {
                 sequence = (FsOptionsSequence *) sequence_SVIDEO_LCD;
             }
            break;
        case Display_Output_DVI:
             if (currentDisplayOutput == Display_Output_LCD) {
                 sequence = (FsOptionsSequence *) sequence_LCD_DVI;
             }
             else if (currentDisplayOutput == Display_Output_SVIDEO) {
                 sequence = (FsOptionsSequence *) sequence_SVIDEO_DVI;
             }
            break;
        case Display_Output_SVIDEO:
             if (currentDisplayOutput == Display_Output_LCD) {
                 sequence = (FsOptionsSequence *) sequence_LCD_SVIDEO;
             }
             else if (currentDisplayOutput == Display_Output_DVI) {
                 sequence = (FsOptionsSequence *) sequence_DVI_SVIDEO;
             }
            break;
        default:
            break;
    }

    if (!sequence) {
        Dmai_err1("Invalid Display Output selected %d\n", *displayOutput);
        return Dmai_EINVAL;
    }

    /* Execute the Required Sequence */
    while (sequence->command != FsOptions_COMMANDS_END) {
        if (setupDevice(sequence->command, sequence->sysFB) < 0) {
            Dmai_err2("Failed to Execute command %s for Output %d\n", 
                      sequence->sysFB, *displayOutput);
            return Dmai_EIO;
        }
    
        sequence++;
    }

    return Dmai_EOK;   
}
